@@ -3,7 +3,7 @@
  * 
  * Written by Ondrej Sury in 2007
  * 
- * Modified a bit my NLnet Labs.
+ * Modified a bit by NLnet Labs.
  * 
  * See the file LICENSE for the license
  */
@@ -30,6 +30,8 @@ usage(int argc, char **argv)
 	printf("       -i - print inserted\n");
 	printf("       -d - print deleted\n");
 	printf("       -c - print changed\n");
+	printf("       -a - print all differences (-i -d -c)\n");
+	printf("       -z - do not sort zones\n");
 }
 
 int 
@@ -41,16 +43,17 @@ main(int argc, char **argv)
 	ldns_status	s;
 	size_t		i      , j;
 	ldns_rr_list   *rrl1, *rrl2;
-	int		rr_cmp    , rr_chg;
+	int		rr_cmp, rr_chg = 0;
 	ldns_rr        *rr1 = NULL, *rr2 = NULL, *rrx = NULL;
 	int		line_nr1 = 0, line_nr2 = 0;
 	size_t		rrc1   , rrc2;
 	size_t		num_ins = 0, num_del = 0, num_chg = 0;
 	int		c;
-	int		opt_deleted = 0, opt_inserted = 0, opt_changed = 0;
+	bool		opt_deleted = false, opt_inserted = false, opt_changed = false;
+        bool		sort = true;
 	char		op = 0;
 
-	while ((c = getopt(argc, argv, "hvdic")) != -1) {
+	while ((c = getopt(argc, argv, "ahvdicz")) != -1) {
 		switch (c) {
 		case 'h':
 			usage(argc, argv);
@@ -63,14 +66,22 @@ main(int argc, char **argv)
 				  ldns_version());
 			exit(EXIT_SUCCESS);
 			break;
+		case 'z':
+			sort = false;
+                        break;
 		case 'd':
-			opt_deleted = 1;
+			opt_deleted = true;
 			break;
 		case 'i':
-			opt_inserted = 1;
+			opt_inserted = true;
 			break;
 		case 'c':
-			opt_changed = 1;
+			opt_changed = true;
+			break;
+		case 'a':
+			opt_deleted = true;
+			opt_inserted = true;
+			opt_changed = true;
 			break;
 		}
 	}
@@ -96,16 +107,13 @@ main(int argc, char **argv)
 	if (s != LDNS_STATUS_OK) {
 		fclose(fp1);
 		fprintf(stderr, "%s: %s at %d\n",
-			fn1,
-			ldns_get_errorstr_by_id(s),
-			line_nr1);
+			   fn1,
+			   ldns_get_errorstr_by_id(s),
+			   line_nr1);
 		exit(EXIT_FAILURE);
 	}
 	fclose(fp1);
 
-	/* Sort first zone */
-	ldns_zone_sort(z1);
-
 	fn2 = argv[1];
 	fp2 = fopen(fn2, "r");
 	if (!fp2) {
@@ -119,22 +127,36 @@ main(int argc, char **argv)
 		ldns_zone_deep_free(z1);
 		fclose(fp2);
 		fprintf(stderr, "%s: %s at %d\n",
-			fn2,
-			ldns_get_errorstr_by_id(s),
-			line_nr2);
+			   fn2,
+			   ldns_get_errorstr_by_id(s),
+			   line_nr2);
 		exit(EXIT_FAILURE);
 	}
 	fclose(fp2);
 
-	/* Sort second zone */
-	ldns_zone_sort(z2);
-
 	rrl1 = ldns_zone_rrs(z1);
 	rrc1 = ldns_rr_list_rr_count(rrl1);
 
 	rrl2 = ldns_zone_rrs(z2);
 	rrc2 = ldns_rr_list_rr_count(rrl2);
 
+        if (sort) {
+		/* canonicalize zone 1 */
+		ldns_rr2canonical(ldns_zone_soa(z1));
+                for (i = 0; i < ldns_rr_list_rr_count(ldns_zone_rrs(z1)); i++) {
+                	ldns_rr2canonical(ldns_rr_list_rr(ldns_zone_rrs(z1), i));
+		}
+                /* sort zone 1 */
+                ldns_zone_sort(z1);
+		/* canonicalize zone 2 */
+		ldns_rr2canonical(ldns_zone_soa(z2));
+                for (i = 0; i < ldns_rr_list_rr_count(ldns_zone_rrs(z2)); i++) {
+                	ldns_rr2canonical(ldns_rr_list_rr(ldns_zone_rrs(z2), i));
+		}
+                /* sort zone 2 */
+                ldns_zone_sort(z2);
+        }
+
 	/*
 	 * Walk through both zones. The previously seen resource record is
 	 * kept (in the variable rrx) so that we can recognize when we are
