@@ -16,9 +16,15 @@
 
 #include <limits.h>
 
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif
+#ifdef HAVE_ARPA_INET_H
 #include <arpa/inet.h>
+#endif
+#ifdef HAVE_NETDB_H
 #include <netdb.h>
+#endif
 #include <time.h>
 #include <sys/time.h>
 
@@ -59,6 +65,7 @@ ldns_lookup_table ldns_rr_classes[] = {
         { LDNS_RR_CLASS_IN, "IN" },
         { LDNS_RR_CLASS_CH, "CH" },
         { LDNS_RR_CLASS_HS, "HS" },
+        { LDNS_RR_CLASS_NONE, "NONE" },
         { LDNS_RR_CLASS_ANY, "ANY" },
         { 0, NULL }
 };
@@ -88,6 +95,130 @@ ldns_lookup_table ldns_opcodes[] = {
         { 0, NULL }
 };
 
+ldns_status
+ldns_pkt_opcode2buffer_str(ldns_buffer *output, ldns_pkt_opcode opcode)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_opcodes, opcode);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "OPCODE%u", opcode);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_pkt_rcode2buffer_str(ldns_buffer *output, ldns_pkt_rcode rcode)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_rcodes, rcode);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "RCODE%u", rcode);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_algorithm2buffer_str(ldns_buffer *output,
+                          ldns_algorithm algorithm)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_algorithms,
+	                                          algorithm);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "ALG%u", algorithm);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_cert_algorithm2buffer_str(ldns_buffer *output,
+                               ldns_cert_algorithm cert_algorithm)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_cert_algorithms,
+	                                          cert_algorithm);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "CERT_ALG%u",
+		                   cert_algorithm);
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_pkt_opcode2str(ldns_pkt_rcode opcode)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(12);
+	str = NULL;
+
+	if (ldns_pkt_opcode2buffer_str(buf, opcode) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_rcode2str(ldns_pkt_rcode rcode)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_pkt_rcode2buffer_str(buf, rcode) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_algorithm2str(ldns_algorithm algorithm)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_algorithm2buffer_str(buf, algorithm)
+	    == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_cert_algorithm2str(ldns_algorithm cert_algorithm)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_cert_algorithm2buffer_str(buf, cert_algorithm)
+	    == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+
 /* do NOT pass compressed data here :p */
 ldns_status
 ldns_rdf2buffer_str_dname(ldns_buffer *output, const ldns_rdf *dname)
@@ -214,7 +345,7 @@ ldns_rdf2buffer_str_str(ldns_buffer *output, const ldns_rdf *rdf)
 	ldns_buffer_printf(output, "\"");
 	for (i = 1; i <= length; ++i) {
 		char ch = (char) data[i];
-		if (isprint(ch)) {
+		if (isprint((int)ch)) {
 			if (ch == '"' || ch == '\\') {
 				ldns_buffer_printf(output, "\\");
 			}
@@ -230,13 +361,28 @@ ldns_rdf2buffer_str_str(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_b64(ldns_buffer *output, const ldns_rdf *rdf)
 {
-	size_t size = b64_ntop_calculate_size(ldns_rdf_size(rdf));
+	size_t size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf));
 	char *b64 = LDNS_XMALLOC(char, size);
-	if (b64_ntop(ldns_rdf_data(rdf), ldns_rdf_size(rdf), b64, size)) {
+	if (ldns_b64_ntop(ldns_rdf_data(rdf), ldns_rdf_size(rdf), b64, size)) {
 		ldns_buffer_printf(output, "%s", b64);
 	}
 	LDNS_FREE(b64);
 	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_rdf2buffer_str_b32_ext(ldns_buffer *output, const ldns_rdf *rdf)
+{
+	size_t size = ldns_b32_ntop_calculate_size(ldns_rdf_size(rdf) - 1);
+	char *b32 = LDNS_XMALLOC(char, size + 1);
+	size = (size_t) ldns_b32_ntop_extended_hex(ldns_rdf_data(rdf) + 1,
+									   ldns_rdf_size(rdf) - 1,
+									   b32, size);
+	if (size > 0) {
+		ldns_buffer_printf(output, "%s", b32);
+	}
+	LDNS_FREE(b32);
+	return ldns_buffer_status(output);
 }	
 
 ldns_status
@@ -257,7 +403,7 @@ ldns_rdf2buffer_str_type(ldns_buffer *output, const ldns_rdf *rdf)
 	const ldns_rr_descriptor *descriptor;
 
 	descriptor = ldns_rr_descript(data);
-	if (descriptor->_name) {
+	if (descriptor && descriptor->_name) {
 		ldns_buffer_printf(output, "%s", descriptor->_name);
 	} else {
 		ldns_buffer_printf(output, "TYPE%u", data);
@@ -268,7 +414,7 @@ ldns_rdf2buffer_str_type(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_class(ldns_buffer *output, const ldns_rdf *rdf)
 {
-        uint8_t data = ldns_rdf_data(rdf)[0];
+	uint16_t data = ldns_read_uint16(ldns_rdf_data(rdf));
 	ldns_lookup_table *lt;
 
  	lt = ldns_lookup_by_id(ldns_rr_classes, (int) data);
@@ -297,8 +443,11 @@ ldns_rdf2buffer_str_cert_alg(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_alg(ldns_buffer *output, const ldns_rdf *rdf)
 {
+	/* don't use algorithm mnemonics in the presentation format
+	   this kind of got sneaked into the rfc's */
         uint8_t data = ldns_rdf_data(rdf)[0];
 /*
+
 	ldns_lookup_table *lt;
 
  	lt = ldns_lookup_by_id(ldns_algorithms, (int) data);
@@ -330,6 +479,88 @@ loc_cm_print(ldns_buffer *output, uint8_t mantissa, uint8_t exponent)
 		ldns_buffer_printf(output, "0");
 }
 
+ldns_status
+ldns_rr_type2buffer_str(ldns_buffer *output, const ldns_rr_type type)
+{
+	const ldns_rr_descriptor *descriptor;
+
+	descriptor = ldns_rr_descript(type);
+
+	if (descriptor && descriptor->_name) {
+		ldns_buffer_printf(output, "%s", descriptor->_name);
+	} else {
+		/* exceptions for pseudotypes */
+		switch (type) {
+			case LDNS_RR_TYPE_IXFR:
+				ldns_buffer_printf(output, "IXFR");
+				break;
+			case LDNS_RR_TYPE_AXFR:
+				ldns_buffer_printf(output, "AXFR");
+				break;
+			case LDNS_RR_TYPE_MAILA:
+				ldns_buffer_printf(output, "MAILA");
+				break;
+			case LDNS_RR_TYPE_MAILB:
+				ldns_buffer_printf(output, "MAILB");
+				break;
+			case LDNS_RR_TYPE_ANY:
+				ldns_buffer_printf(output, "ANY");
+				break;
+			default:
+				ldns_buffer_printf(output, "TYPE%u", type);
+		}
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_rr_type2str(const ldns_rr_type type)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_rr_type2buffer_str(buf, type) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+
+ldns_status
+ldns_rr_class2buffer_str(ldns_buffer *output,
+                         const ldns_rr_class klass)
+{
+	ldns_lookup_table *lt;
+
+	lt = ldns_lookup_by_id(ldns_rr_classes, klass);
+	if (lt) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "CLASS%d", klass);
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_rr_class2str(const ldns_rr_class klass)
+{
+	ldns_buffer *buf;
+	char *str;
+
+	str = NULL;
+	buf = ldns_buffer_new(10);
+	if (ldns_rr_class2buffer_str(buf, klass) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+	ldns_buffer_free(buf);
+	return str;
+}
+
 ldns_status
 ldns_rdf2buffer_str_loc(ldns_buffer *output, const ldns_rdf *rdf)
 {
@@ -340,7 +571,7 @@ ldns_rdf2buffer_str_loc(ldns_buffer *output, const ldns_rdf *rdf)
 	uint8_t vertical_precision;
 	uint32_t longitude;
 	uint32_t latitude;
-	uint32_t altitude;
+	long altitude;
 	char northerness;
 	char easterness;
 	uint32_t h;
@@ -443,7 +674,9 @@ ldns_rdf2buffer_str_wks(ldns_buffer *output, const ldns_rdf *rdf)
 		ldns_buffer_printf(output, "%u ", protocol_nr);
 	}
 
+#ifdef HAVE_ENDPROTOENT
 	endprotoent();
+#endif
 	
 	for (current_service = 0; 
 	     current_service < ldns_rdf_size(rdf) * 7; current_service++) {
@@ -455,7 +688,9 @@ ldns_rdf2buffer_str_wks(ldns_buffer *output, const ldns_rdf *rdf)
 			} else {
 				ldns_buffer_printf(output, "%u ", current_service);
 			}
+#ifdef HAVE_ENDSERVENT
 			endservent();
+#endif
 		}
 	}
 	return ldns_buffer_status(output);
@@ -485,11 +720,11 @@ ldns_rdf2buffer_str_nsec(ldns_buffer *output, const ldns_rdf *rdf)
 				type = 256 * (uint16_t) window_block_nr + bit_pos;
 				descriptor = ldns_rr_descript(type);
 
-				if (descriptor->_name) {
+				if (descriptor && descriptor->_name) {
 					ldns_buffer_printf(output, "%s ", 
 							descriptor->_name);
 				} else {
-					ldns_buffer_printf(output, "TYPE%d ", type);
+					ldns_buffer_printf(output, "TYPE%u ", type);
 				}
 			}
 		}
@@ -500,6 +735,32 @@ ldns_rdf2buffer_str_nsec(ldns_buffer *output, const ldns_rdf *rdf)
 	return ldns_buffer_status(output);
 }
 
+ldns_status
+ldns_rdf2buffer_str_nsec3_salt(ldns_buffer *output, const ldns_rdf *rdf)
+{
+	uint8_t salt_length;
+	uint8_t salt_pos;
+
+	uint8_t *data = ldns_rdf_data(rdf);
+	size_t pos;
+
+	salt_length = data[0];
+	/* todo: length check needed/possible? */
+	/* from now there are variable length entries so remember pos */
+	pos = 1;
+	if (salt_length == 0) {
+		ldns_buffer_printf(output, "- ");
+	} else {
+		for (salt_pos = 0; salt_pos < salt_length; salt_pos++) {
+			ldns_buffer_printf(output, "%02x", data[1 + salt_pos]);
+			pos++;
+		}
+		ldns_buffer_printf(output, " ");
+	}
+	
+	return ldns_buffer_status(output);
+}
+
 ldns_status
 ldns_rdf2buffer_str_period(ldns_buffer *output, const ldns_rdf *rdf)
 {
@@ -602,13 +863,15 @@ ldns_status
 ldns_rdf2buffer_str_int16_data(ldns_buffer *output, const ldns_rdf *rdf)
 {
 	/* Subtract the size (2) of the number that specifies the length */
-	size_t size = b64_ntop_calculate_size(ldns_rdf_size(rdf) - 2);
+	size_t size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf) - 2);
 	char *b64 = LDNS_XMALLOC(char, size);
 
 	ldns_buffer_printf(output, "%u ", ldns_rdf_size(rdf) - 2);
 	
 	if (ldns_rdf_size(rdf) > 2 &&
-	    b64_ntop(ldns_rdf_data(rdf) + 2, ldns_rdf_size(rdf) - 2, b64, size)) {
+	    ldns_b64_ntop(ldns_rdf_data(rdf) + 2,
+				   ldns_rdf_size(rdf) - 2,
+				   b64, size)) {
 		ldns_buffer_printf(output, "%s", b64);
 	}
 	LDNS_FREE(b64);
@@ -626,7 +889,7 @@ ldns_rdf2buffer_str_ipseckey(ldns_buffer *output, const ldns_rdf *rdf)
 	uint8_t gateway_type;
 	uint8_t algorithm;
 	
-	ldns_rdf *gateway;
+	ldns_rdf *gateway = NULL;
 	uint8_t *gateway_data;
 	
 	size_t public_key_size;
@@ -729,6 +992,9 @@ ldns_rdf2buffer_str(ldns_buffer *buffer, const ldns_rdf *rdf)
 		case LDNS_RDF_TYPE_APL:
 			res = ldns_rdf2buffer_str_apl(buffer, rdf);
 			break;
+		case LDNS_RDF_TYPE_B32_EXT:
+			res = ldns_rdf2buffer_str_b32_ext(buffer, rdf);
+			break;
 		case LDNS_RDF_TYPE_B64:
 			res = ldns_rdf2buffer_str_b64(buffer, rdf);
 			break;
@@ -738,6 +1004,9 @@ ldns_rdf2buffer_str(ldns_buffer *buffer, const ldns_rdf *rdf)
 		case LDNS_RDF_TYPE_NSEC: 
 			res = ldns_rdf2buffer_str_nsec(buffer, rdf);
 			break;
+		case LDNS_RDF_TYPE_NSEC3_SALT:
+			res = ldns_rdf2buffer_str_nsec3_salt(buffer, rdf);
+			break;
 		case LDNS_RDF_TYPE_TYPE: 
 			res = ldns_rdf2buffer_str_type(buffer, rdf);
 			break;
@@ -775,6 +1044,9 @@ ldns_rdf2buffer_str(ldns_buffer *buffer, const ldns_rdf *rdf)
 		case LDNS_RDF_TYPE_INT16_DATA:
 			res = ldns_rdf2buffer_str_int16_data(buffer, rdf);
 			break;
+		case LDNS_RDF_TYPE_NSEC3_NEXT_OWNER:
+			res = ldns_rdf2buffer_str_b32_ext(buffer, rdf);
+			break;
 		}
 	} else {
 		ldns_buffer_printf(buffer, "(null) ");
@@ -785,11 +1057,8 @@ ldns_rdf2buffer_str(ldns_buffer *buffer, const ldns_rdf *rdf)
 ldns_status
 ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 {
-	uint16_t i;
+	uint16_t i, flags;
 	ldns_status status = LDNS_STATUS_OK;
-	ldns_lookup_table *lt;
-	const ldns_rr_descriptor *descriptor;
-
 	if (!rr) {
 		ldns_buffer_printf(output, "(null)\n");
 	} else {
@@ -807,36 +1076,12 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 		if (ldns_rr_rd_count(rr) > 0) {
 			ldns_buffer_printf(output, "\t%d", ldns_rr_ttl(rr));
 		}
-		
-		lt = ldns_lookup_by_id(ldns_rr_classes, ldns_rr_get_class(rr));
-		if (lt) {
-			ldns_buffer_printf(output, "\t%s\t", lt->name);
-		} else {
-			ldns_buffer_printf(output, "\tCLASS%d\t", 
-					ldns_rr_get_class(rr));
-		}
 
-		descriptor = ldns_rr_descript(ldns_rr_get_type(rr));
+		ldns_buffer_printf(output, "\t");
+		ldns_rr_class2buffer_str(output, ldns_rr_get_class(rr));
+		ldns_buffer_printf(output, "\t");
 
-		if (descriptor->_name) {
-			ldns_buffer_printf(output, "%s", descriptor->_name);
-		} else {
-			/* exceptions for qtype */
-			if (ldns_rr_get_type(rr) == 251) {
-				ldns_buffer_printf(output, "IXFR ");
-			} else if (ldns_rr_get_type(rr) == 252) {
-				ldns_buffer_printf(output, "AXFR ");
-			} else if (ldns_rr_get_type(rr) == 253) {
-				ldns_buffer_printf(output, "MAILB ");
-			} else if (ldns_rr_get_type(rr) == 254) {
-				ldns_buffer_printf(output, "MAILA ");
-			} else if (ldns_rr_get_type(rr) == 255) {
-				ldns_buffer_printf(output, "ANY ");
-			} else {
-				ldns_buffer_printf(output, "TYPE%d\t", 
-						ldns_rr_get_type(rr));
-			}
-		}
+		ldns_rr_type2buffer_str(output, ldns_rr_get_type(rr));
 		
 		if (ldns_rr_rd_count(rr) > 0) {
 			ldns_buffer_printf(output, "\t");
@@ -845,7 +1090,7 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 			status = ldns_rdf2buffer_str(output, ldns_rr_rdf(rr, i));
 			if (i < ldns_rr_rd_count(rr) - 1) {
 				ldns_buffer_printf(output, " ");
-			} 
+			}
 		}
 		/* per RR special comments - handy for DNSSEC types */
 		/* check to prevent question sec. rr from
@@ -854,14 +1099,15 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 			switch (ldns_rr_get_type(rr)) {
 				case LDNS_RR_TYPE_DNSKEY:
 #ifdef HAVE_SSL
-					if (ldns_rdf2native_int16(ldns_rr_rdf(rr, 0)) == 256) {
+					flags = ldns_rdf2native_int16(ldns_rr_rdf(rr, 0));
+					if (flags == 256 || flags == 384) {
 						ldns_buffer_printf(output, 
 								" ;{id = %d (zsk), size = %db}", 
 								ldns_calc_keytag(rr),
 								ldns_rr_dnskey_key_size(rr)); 
 						break;
 					} 
-					if (ldns_rdf2native_int16(ldns_rr_rdf(rr, 0)) == 257) {
+					if (flags == 256 || flags == 385) {
 						ldns_buffer_printf(output, 
 								" ;{id = %d (ksk), size = %db}", 
 								ldns_calc_keytag(rr),
@@ -876,7 +1122,16 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 				case LDNS_RR_TYPE_RRSIG:
 					ldns_buffer_printf(output, " ;{id = %d}", 
 							ldns_rdf2native_int16(ldns_rr_rdf(rr, 6)));
-							break;
+					break;
+				case LDNS_RR_TYPE_DS:
+					{
+						uint8_t *data = ldns_rdf_data(ldns_rr_rdf(rr, 3));
+						size_t len = ldns_rdf_size(ldns_rr_rdf(rr, 3));
+						char *babble = ldns_bubblebabble(data, len);
+						ldns_buffer_printf(output, " ; %s", babble);
+						LDNS_FREE(babble);
+					}
+					break;
 				default:
 					break;
 
@@ -1027,6 +1282,10 @@ ldns_pkt2buffer_str(ldns_buffer *output, const ldns_pkt *pkt)
 			if (ldns_pkt_edns_do(pkt)) {
 				ldns_buffer_printf(output, " do");
 			}
+			if (ldns_pkt_edns_extended_rcode(pkt)) {
+				ldns_buffer_printf(output, " ; ext-rcode: %d",
+					ldns_pkt_edns_extended_rcode(pkt));
+			}
 			ldns_buffer_printf(output, " ; udp: %u\n",
 					   ldns_pkt_edns_udp_size(pkt));
 			
@@ -1060,15 +1319,37 @@ ldns_pkt2buffer_str(ldns_buffer *output, const ldns_pkt *pkt)
 	return status;
 }
 
+static ldns_status
+ldns_hmac_key2buffer_str(ldns_buffer *output, const ldns_key *k)
+{
+	ldns_status status;
+	size_t i;
+	ldns_rdf *b64_bignum;
+	
+	ldns_buffer_printf(output, "Key: ");
+
+ 	i = ldns_key_hmac_size(k);
+	b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, ldns_key_hmac_key(k));
+	status = ldns_rdf2buffer_str(output, b64_bignum);
+	ldns_rdf_deep_free(b64_bignum);
+	ldns_buffer_printf(output, "\n");
+	return status;
+}
+
+
 ldns_status
 ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 {
 	ldns_status status = LDNS_STATUS_OK;
 	unsigned char  *bignum;
+	
 #ifdef HAVE_SSL
 	/* not used when ssl is not defined */
 	ldns_rdf *b64_bignum = NULL;
 	uint16_t i;
+	
+	RSA *rsa;
+	DSA *dsa;
 #endif /* HAVE_SSL */
 
 	if (!k) {
@@ -1084,20 +1365,69 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 #ifdef HAVE_SSL
 		switch(ldns_key_algorithm(k)) {
 			case LDNS_SIGN_RSASHA1:
+			case LDNS_SIGN_RSASHA1_NSEC3:
+			case LDNS_SIGN_RSASHA256:
+			case LDNS_SIGN_RSASHA256_NSEC3:
+			case LDNS_SIGN_RSASHA512:
+			case LDNS_SIGN_RSASHA512_NSEC3:
 			case LDNS_SIGN_RSAMD5:
 				/* copied by looking at dnssec-keygen output */
 				/* header */
+				rsa = ldns_key_rsa_key(k);
+
 				ldns_buffer_printf(output,"Private-key-format: v1.2\n");
-				if (ldns_key_algorithm(k) == LDNS_SIGN_RSAMD5) {
-					ldns_buffer_printf(output,"Algorithm: 1 (RSA)\n");
-				} else if (ldns_key_algorithm(k) == LDNS_SIGN_RSASHA1) {
-					ldns_buffer_printf(output,"Algorithm: 5 (RSASHA1)\n");
+				switch(ldns_key_algorithm(k)) {
+				case LDNS_SIGN_RSAMD5:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSA)\n",
+								    LDNS_RSAMD5);
+					break;
+				case LDNS_SIGN_RSASHA1:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA1)\n",
+								    LDNS_RSASHA1);
+					break;
+				case LDNS_SIGN_RSASHA1_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA1_NSEC3)\n",
+								    LDNS_RSASHA1_NSEC3);
+					break;
+#ifdef USE_SHA2
+				case LDNS_SIGN_RSASHA256:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA256)\n",
+								    LDNS_RSASHA256);
+					break;
+				case LDNS_SIGN_RSASHA256_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA256_NSEC3)\n",
+								    LDNS_RSASHA256_NSEC3);
+					break;
+				case LDNS_SIGN_RSASHA512:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA512)\n",
+								    LDNS_RSASHA512);
+					break;
+				case LDNS_SIGN_RSASHA512_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA512_NSEC3)\n",
+								    LDNS_RSASHA512_NSEC3);
+					break;
+#endif
+				default:
+					fprintf(stderr, "Warning: unknown signature ");
+					fprintf(stderr,
+						   "algorithm type %u\n", 
+						   ldns_key_algorithm(k));
+					break;
 				}
+				
+
 
 				/* print to buf, convert to bin, convert to b64,
 				 * print to buf */
 				ldns_buffer_printf(output, "Modulus: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->n, bignum);
+				i = (uint16_t)BN_bn2bin(rsa->n, bignum);
 				if (i > LDNS_MAX_KEYLEN) {
 					goto error;
 				}
@@ -1107,9 +1437,8 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 				}
 				ldns_rdf_deep_free(b64_bignum);
 				ldns_buffer_printf(output, "\n"); 
-				
 				ldns_buffer_printf(output, "PublicExponent: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->e, bignum);
+				i = (uint16_t)BN_bn2bin(rsa->e, bignum);
 				if (i > LDNS_MAX_KEYLEN) {
 					goto error;
 				}
@@ -1121,155 +1450,212 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "PrivateExponent: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->d, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->d) {
+					i = (uint16_t)BN_bn2bin(rsa->d, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
-
+				
 				ldns_buffer_printf(output, "Prime1: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->p, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->p) {
+					i = (uint16_t)BN_bn2bin(rsa->p, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Prime2: ");
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->q, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->q) {
+					i = (uint16_t)BN_bn2bin(rsa->q, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Exponent1: ");
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->dmp1, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->dmp1) {
+					i = (uint16_t)BN_bn2bin(rsa->dmp1, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Exponent2: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->dmq1, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->dmq1) {
+					i = (uint16_t)BN_bn2bin(rsa->dmq1, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Coefficient: "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_rsa_key(k)->iqmp, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (rsa->iqmp) {
+					i = (uint16_t)BN_bn2bin(rsa->iqmp, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					ldns_buffer_printf(output, "(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
+				
+				RSA_free(rsa);
 				break;
 			case LDNS_SIGN_DSA:
+			case LDNS_SIGN_DSA_NSEC3:
+				dsa = ldns_key_dsa_key(k);
+			
 				ldns_buffer_printf(output,"Private-key-format: v1.2\n");
-				ldns_buffer_printf(output,"Algorithm: 3 (DSA)\n");
+				if (ldns_key_algorithm(k) == LDNS_SIGN_DSA) {
+					ldns_buffer_printf(output,"Algorithm: 3 (DSA)\n");
+				} else if (ldns_key_algorithm(k) == LDNS_SIGN_DSA_NSEC3) {
+					ldns_buffer_printf(output,"Algorithm: 6 (DSA_NSEC3)\n");
+				}
 
 				/* print to buf, convert to bin, convert to b64,
 				 * print to buf */
 				ldns_buffer_printf(output, "Prime(p): "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_dsa_key(k)->p, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (dsa->p) {
+					i = (uint16_t)BN_bn2bin(dsa->p, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					printf("(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Subprime(q): "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_dsa_key(k)->q, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (dsa->q) {
+					i = (uint16_t)BN_bn2bin(dsa->q, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					printf("(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Base(g): "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_dsa_key(k)->g, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (dsa->g) {
+					i = (uint16_t)BN_bn2bin(dsa->g, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					printf("(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Private_value(x): "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_dsa_key(k)->priv_key, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (dsa->priv_key) {
+					i = (uint16_t)BN_bn2bin(dsa->priv_key, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					printf("(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 
 				ldns_buffer_printf(output, "Public_value(y): "); 
-				i = (uint16_t)BN_bn2bin(ldns_key_dsa_key(k)->pub_key, bignum);
-				if (i > LDNS_MAX_KEYLEN) {
-					goto error;
+				if (dsa->pub_key) {
+					i = (uint16_t)BN_bn2bin(dsa->pub_key, bignum);
+					if (i > LDNS_MAX_KEYLEN) {
+						goto error;
+					}
+					b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
+					if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
+						goto error;
+					}
+					ldns_rdf_deep_free(b64_bignum);
+					ldns_buffer_printf(output, "\n"); 
+				} else {
+					printf("(Not available)\n");
 				}
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, bignum);
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
 				break;
 			case LDNS_SIGN_HMACMD5:
-				/* is the filefmt specified for TSIG.. don't know */
-		                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
-		                ldns_buffer_printf(output, "Algorithm: 157 (HMAC_MD5)\n");
-		                ldns_buffer_printf(output, "Key: ");
-		                i = ldns_key_hmac_size(k);
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, ldns_key_hmac_key(k));
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
+				/* there's not much of a format defined for TSIG */
+				/* It's just a binary blob, Same for all algorithms */
+                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+                ldns_buffer_printf(output, "Algorithm: 157 (HMAC_MD5)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
+			case LDNS_SIGN_HMACSHA1:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 158 (HMAC_SHA1)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
+			case LDNS_SIGN_HMACSHA256:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 158 (HMAC_SHA256)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
 				break;
 		}
 #endif /* HAVE_SSL */
@@ -1296,7 +1682,7 @@ error:
  * Zero terminate the buffer and fix it to the size of the string.
  */
 char *
-buffer2str(ldns_buffer *buffer)
+ldns_buffer2str(ldns_buffer *buffer)
 {
 	char *tmp_str;
 	char *str;
@@ -1328,7 +1714,7 @@ ldns_rdf2str(const ldns_rdf *rdf)
 
 	if (ldns_rdf2buffer_str(tmp_buffer, rdf) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 	
 	ldns_buffer_free(tmp_buffer);
@@ -1343,9 +1729,8 @@ ldns_rr2str(const ldns_rr *rr)
 
 	if (ldns_rr2buffer_str(tmp_buffer, rr) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
-	
 	ldns_buffer_free(tmp_buffer);
 	return result;
 }
@@ -1354,11 +1739,11 @@ char *
 ldns_pkt2str(const ldns_pkt *pkt)
 {
 	char *result = NULL;
-	ldns_buffer *tmp_buffer = ldns_buffer_new(LDNS_MIN_BUFLEN);
+	ldns_buffer *tmp_buffer = ldns_buffer_new(LDNS_MAX_PACKETLEN);
 
 	if (ldns_pkt2buffer_str(tmp_buffer, pkt) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 
 	ldns_buffer_free(tmp_buffer);
@@ -1372,7 +1757,7 @@ ldns_key2str(const ldns_key *k)
 	ldns_buffer *tmp_buffer = ldns_buffer_new(LDNS_MIN_BUFLEN);
 	if (ldns_key2buffer_str(tmp_buffer, k) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 	ldns_buffer_free(tmp_buffer);
 	return result;
@@ -1392,7 +1777,7 @@ ldns_rr_list2str(const ldns_rr_list *list)
 	}
 
 	/* export and return string, destroy rest */
-	result = buffer2str(tmp_buffer);
+	result = ldns_buffer2str(tmp_buffer);
 	ldns_buffer_free(tmp_buffer);
 	return result;
 }
