@@ -6,7 +6,7 @@
  * handling
  *
  * (c) NLnet Labs, 2004-2006
- * 
+ *
  * See the file LICENSE for the license
  */
 
@@ -16,18 +16,29 @@
 
 #ifdef HAVE_SSL
 #include <openssl/ssl.h>
+#include <openssl/engine.h>
 #include <openssl/rand.h>
 #endif /* HAVE_SSL */
 
 ldns_lookup_table ldns_signing_algorithms[] = {
         { LDNS_SIGN_RSAMD5, "RSAMD5" },
         { LDNS_SIGN_RSASHA1, "RSASHA1" },
-        { LDNS_SIGN_DSA, "DSAMD5" },
+        { LDNS_SIGN_RSASHA1_NSEC3, "RSASHA1_NSEC3" },
+#ifdef USE_SHA2
+        { LDNS_SIGN_RSASHA256, "RSASHA256" },
+        { LDNS_SIGN_RSASHA256_NSEC3, "RSASHA256_NSEC3" },
+        { LDNS_SIGN_RSASHA512, "RSASHA512" },
+        { LDNS_SIGN_RSASHA512_NSEC3, "RSASHA512_NSEC3" },
+#endif
+        { LDNS_SIGN_DSA, "DSA" },
+        { LDNS_SIGN_DSA_NSEC3, "DSA_NSEC3" },
         { LDNS_SIGN_HMACMD5, "hmac-md5.sig-alg.reg.int" },
+        { LDNS_SIGN_HMACSHA1, "hmac-sha1" },
+        { LDNS_SIGN_HMACSHA256, "hmac-sha256" },
         { 0, NULL }
 };
 
-#ifdef HAVE_SSL 
+#ifdef HAVE_SSL
 ldns_key_list *
 ldns_key_list_new()
 {
@@ -51,25 +62,43 @@ ldns_key_new()
 		return NULL;
 	} else {
 		/* some defaults - not sure wether to do this */
+		ldns_key_set_use(newkey, true);
 		ldns_key_set_flags(newkey, LDNS_KEY_ZONE_KEY);
 		ldns_key_set_origttl(newkey, 0);
 		ldns_key_set_keytag(newkey, 0);
 		ldns_key_set_inception(newkey, 0);
 		ldns_key_set_expiration(newkey, 0);
 		ldns_key_set_pubkey_owner(newkey, NULL);
-		ldns_key_set_rsa_key(newkey, NULL);
-		ldns_key_set_dsa_key(newkey, NULL);
 		ldns_key_set_hmac_key(newkey, NULL);
 		return newkey;
 	}
+	return NULL;
 }
 
-ldns_status 
+ldns_status
 ldns_key_new_frm_fp(ldns_key **k, FILE *fp)
 {
 	return ldns_key_new_frm_fp_l(k, fp, NULL);
 }
 
+#ifdef HAVE_SSL
+ldns_status
+ldns_key_new_frm_engine(ldns_key **key, ENGINE *e, char *key_id, ldns_algorithm alg)
+{
+	ldns_key *k;
+
+	k = ldns_key_new();
+	k->_key.key = ENGINE_load_private_key(e, key_id, UI_OpenSSL(), NULL);
+	ldns_key_set_algorithm(k, alg);
+	if (!k->_key.key) {
+		return LDNS_STATUS_ENGINE_KEY_NOT_LOADED;
+	} else {
+		*key = k;
+		return LDNS_STATUS_OK;
+	}
+}
+#endif
+
 ldns_status
 ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 {
@@ -77,6 +106,8 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 	char *d;
 	ldns_signing_algorithm alg;
 	ldns_rr *key_rr;
+	RSA *rsa;
+	DSA *dsa;
 	unsigned char *hmac;
 	size_t hmac_size;
 
@@ -86,9 +117,9 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 	if (!k || !d) {
 		return LDNS_STATUS_MEM_ERR;
 	}
-	
+
 	alg = 0;
-	
+
 	/* the file is highly structured. Do this in sequence */
 	/* RSA:
 	 * Private-key-format: v1.2
@@ -107,7 +138,7 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 
 	/* get the algorithm type, our file function strip ( ) so there are
 	 * not in the return string! */
-	if (ldns_fget_keyword_data_l(fp, "Algorithm", ": ", d, "\n", 
+	if (ldns_fget_keyword_data_l(fp, "Algorithm", ": ", d, "\n",
 				LDNS_MAX_LINELEN, line_nr) == -1) {
 		/* no alg information */
 		return LDNS_STATUS_SYNTAX_ALG_ERR;
@@ -116,31 +147,94 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 	if (strncmp(d, "1 RSA", 2) == 0) {
 		alg = LDNS_SIGN_RSAMD5;
 	}
+	if (strncmp(d, "2 DH", 2) == 0) {
+		alg = LDNS_DH;
+	}
 	if (strncmp(d, "3 DSA", 2) == 0) {
-		alg = LDNS_SIGN_DSA; 
+		alg = LDNS_SIGN_DSA;
+	}
+	if (strncmp(d, "4 ECC", 2) == 0) {
+		alg = LDNS_ECC;
 	}
 	if (strncmp(d, "5 RSASHA1", 2) == 0) {
 		alg = LDNS_SIGN_RSASHA1;
 	}
+	if (strncmp(d, "6 DSA", 2) == 0) {
+		alg = LDNS_DSA_NSEC3;
+	}
+	if (strncmp(d, "7 RSASHA1", 2) == 0) {
+		alg = LDNS_RSASHA1_NSEC3;
+	}
+
+	if (strncmp(d, "8 RSASHA256", 2) == 0) {
+#ifdef USE_SHA2
+		alg = LDNS_SIGN_RSASHA256;
+#else
+		fprintf(stderr, "Warning: SHA256 not compiled into this ");
+		fprintf(stderr, "version of ldns\n");
+#endif
+	}
+	if (strncmp(d, "9 RSASHA256", 2) == 0) {
+#ifdef USE_SHA2
+		alg = LDNS_SIGN_RSASHA256_NSEC3;
+#else
+		fprintf(stderr, "Warning: SHA256 not compiled into this ");
+		fprintf(stderr, "version of ldns\n");
+#endif
+	}
+	if (strncmp(d, "10 RSASHA512", 3) == 0) {
+#ifdef USE_SHA2
+		alg = LDNS_SIGN_RSASHA512;
+#else
+		fprintf(stderr, "Warning: SHA512 not compiled into this ");
+		fprintf(stderr, "version of ldns\n");
+#endif
+	}
+	if (strncmp(d, "11 RSASHA512", 3) == 0) {
+#ifdef USE_SHA2
+		alg = LDNS_SIGN_RSASHA512_NSEC3;
+#else
+		fprintf(stderr, "Warning: SHA512 not compiled into this ");
+		fprintf(stderr, "version of ldns\n");
+#endif
+	}
 	if (strncmp(d, "157 HMAC-MD5", 4) == 0) {
 		alg = LDNS_SIGN_HMACMD5;
 	}
+	if (strncmp(d, "158 HMAC-SHA1", 4) == 0) {
+		alg = LDNS_SIGN_HMACSHA1;
+	}
+	if (strncmp(d, "159 HMAC-SHA256", 4) == 0) {
+		alg = LDNS_SIGN_HMACSHA256;
+	}
 
 	LDNS_FREE(d);
 
 	switch(alg) {
 		case LDNS_SIGN_RSAMD5:
 		case LDNS_SIGN_RSASHA1:
-
+		case LDNS_RSASHA1_NSEC3:
+#ifdef USE_SHA2
+		case LDNS_SIGN_RSASHA256:
+		case LDNS_SIGN_RSASHA256_NSEC3:
+		case LDNS_SIGN_RSASHA512:
+		case LDNS_SIGN_RSASHA512_NSEC3:
+#endif
 			ldns_key_set_algorithm(k, alg);
-			ldns_key_set_rsa_key(k, ldns_key_new_frm_fp_rsa_l(fp, line_nr));
-
+			rsa = ldns_key_new_frm_fp_rsa_l(fp, line_nr);
+			ldns_key_set_rsa_key(k, rsa);
+			RSA_free(rsa);
 			break;
 		case LDNS_SIGN_DSA:
+		case LDNS_DSA_NSEC3:
 			ldns_key_set_algorithm(k, alg);
-			ldns_key_set_dsa_key(k, ldns_key_new_frm_fp_dsa_l(fp, line_nr));
+			dsa = ldns_key_new_frm_fp_dsa_l(fp, line_nr);
+			ldns_key_set_dsa_key(k, dsa);
+			DSA_free(dsa);
 			break;
 		case LDNS_SIGN_HMACMD5:
+		case LDNS_SIGN_HMACSHA1:
+		case LDNS_SIGN_HMACSHA256:
 			ldns_key_set_algorithm(k, alg);
 			hmac = ldns_key_new_frm_fp_hmac_l(fp, line_nr, &hmac_size);
 			ldns_key_set_hmac_size(k, hmac_size);
@@ -155,6 +249,7 @@ ldns_key_new_frm_fp_l(ldns_key **key, FILE *fp, int *line_nr)
 	key_rr = ldns_key2rr(k);
 	ldns_key_set_keytag(k, ldns_calc_keytag(key_rr));
 	ldns_rr_free(key_rr);
+
 	if (key) {
 		*key = k;
 		return LDNS_STATUS_OK;
@@ -172,14 +267,14 @@ RSA *
 ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 {
 	/* we parse
- 	 * Modulus: 
- 	 * PublicExponent: 
- 	 * PrivateExponent: 
- 	 * Prime1: 
- 	 * Prime2: 
- 	 * Exponent1: 
- 	 * Exponent2: 
- 	 * Coefficient: 
+ 	 * Modulus:
+ 	 * PublicExponent:
+ 	 * PrivateExponent:
+ 	 * Prime1:
+ 	 * Prime2:
+ 	 * Exponent1:
+ 	 * Exponent2:
+ 	 * Coefficient:
 	 *
 	 * man 3 RSA:
 	 *
@@ -209,14 +304,14 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	}
 
 	/* I could use functions again, but that seems an overkill,
-	 * allthough this also looks tedious 
+	 * allthough this also looks tedious
 	 */
 
 	/* Modules, rsa->n */
 	if (ldns_fget_keyword_data_l(f, "Modulus", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->n = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->n) {
 		goto error;
@@ -226,7 +321,7 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "PublicExponent", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->e = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->e) {
 		goto error;
@@ -236,7 +331,7 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "PrivateExponent", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->d = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->d) {
 		goto error;
@@ -246,17 +341,17 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Prime1", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->p = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->p) {
 		goto error;
 	}
-	
+
 	/* Prime2, rsa->q */
 	if (ldns_fget_keyword_data_l(f, "Prime2", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->q = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->q) {
 		goto error;
@@ -266,17 +361,17 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Exponent1", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->dmp1 = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->dmp1) {
 		goto error;
 	}
-	
+
 	/* Exponent2, rsa->dmq1 */
 	if (ldns_fget_keyword_data_l(f, "Exponent2", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->dmq1 = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->dmq1) {
 		goto error;
@@ -286,12 +381,12 @@ ldns_key_new_frm_fp_rsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Coefficient", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	rsa->iqmp = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!rsa->iqmp) {
 		goto error;
 	}
-	
+
 	LDNS_FREE(buf);
 	LDNS_FREE(d);
 	return rsa;
@@ -331,7 +426,7 @@ ldns_key_new_frm_fp_dsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Primep", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	dsa->p = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!dsa->p) {
 		goto error;
@@ -341,7 +436,7 @@ ldns_key_new_frm_fp_dsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Subprimeq", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	dsa->q = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!dsa->q) {
 		goto error;
@@ -351,7 +446,7 @@ ldns_key_new_frm_fp_dsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Baseg", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	dsa->g = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!dsa->g) {
 		goto error;
@@ -361,7 +456,7 @@ ldns_key_new_frm_fp_dsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Private_valuex", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	dsa->priv_key = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!dsa->priv_key) {
 		goto error;
@@ -371,7 +466,7 @@ ldns_key_new_frm_fp_dsa_l(FILE *f, int *line_nr)
 	if (ldns_fget_keyword_data_l(f, "Public_valuey", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 	dsa->pub_key = BN_bin2bn((const char unsigned*)buf, i, NULL);
 	if (!dsa->pub_key) {
 		goto error;
@@ -405,15 +500,15 @@ ldns_key_new_frm_fp_hmac_l(FILE *f, int *line_nr, size_t *hmac_size)
 
 	d = LDNS_XMALLOC(char, LDNS_MAX_LINELEN);
 	buf = LDNS_XMALLOC(unsigned char, LDNS_MAX_LINELEN);
-	
+
 	if (ldns_fget_keyword_data_l(f, "Key", ": ", d, "\n", LDNS_MAX_LINELEN, line_nr) == -1) {
 		goto error;
 	}
-	i = b64_pton((const char*)d, buf, b64_ntop_calculate_size(strlen(d)));
+	i = ldns_b64_pton((const char*)d, buf, ldns_b64_ntop_calculate_size(strlen(d)));
 
 	*hmac_size = i;
 	return buf;
-	
+
 	error:
 	LDNS_FREE(d);
 	LDNS_FREE(buf);
@@ -441,13 +536,20 @@ ldns_key_new_frm_algorithm(ldns_signing_algorithm alg, uint16_t size)
 	switch(alg) {
 		case LDNS_SIGN_RSAMD5:
 		case LDNS_SIGN_RSASHA1:
-			r = RSA_generate_key((int)size, RSA_3, NULL, NULL);
+		case LDNS_SIGN_RSASHA1_NSEC3:
+		case LDNS_SIGN_RSASHA256:
+		case LDNS_SIGN_RSASHA256_NSEC3:
+		case LDNS_SIGN_RSASHA512:
+		case LDNS_SIGN_RSASHA512_NSEC3:
+			r = RSA_generate_key((int)size, RSA_F4, NULL, NULL);
 			if (RSA_check_key(r) != 1) {
 				return NULL;
 			}
+
 			ldns_key_set_rsa_key(k, r);
 			break;
 		case LDNS_SIGN_DSA:
+		case LDNS_SIGN_DSA_NSEC3:
 			d = DSA_generate_parameters((int)size, NULL, 0, NULL, NULL, NULL, NULL);
 			if (!d) {
 				return NULL;
@@ -458,11 +560,12 @@ ldns_key_new_frm_algorithm(ldns_signing_algorithm alg, uint16_t size)
 			ldns_key_set_dsa_key(k, d);
 			break;
 		case LDNS_SIGN_HMACMD5:
-			k->_key.rsa = NULL;
-			k->_key.dsa = NULL;
+		case LDNS_SIGN_HMACSHA1:
+		case LDNS_SIGN_HMACSHA256:
+			k->_key.key = NULL;
 			size = size / 8;
 			ldns_key_set_hmac_size(k, size);
-			 
+
 			hmac = LDNS_XMALLOC(unsigned char, size);
 #ifdef HAVE_SSL
 			if (RAND_bytes(hmac, size) != 1) {
@@ -482,7 +585,7 @@ ldns_key_new_frm_algorithm(ldns_signing_algorithm alg, uint16_t size)
 			}
 #endif /* HAVE_SSL */
 			ldns_key_set_hmac_key(k, hmac);
-			
+
 			ldns_key_set_flags(k, 0);
 			break;
 	}
@@ -504,7 +607,7 @@ ldns_key_print(FILE *output, const ldns_key *k)
 
 
 void
-ldns_key_set_algorithm(ldns_key *k, ldns_signing_algorithm l) 
+ldns_key_set_algorithm(ldns_key *k, ldns_signing_algorithm l)
 {
 	k->_alg = l;
 }
@@ -515,16 +618,26 @@ ldns_key_set_flags(ldns_key *k, uint16_t f)
 	k->_extra.dnssec.flags = f;
 }
 
+void
+ldns_key_set_evp_key(ldns_key *k, EVP_PKEY *e)
+{
+	k->_key.key = e;
+}
+
 void
 ldns_key_set_rsa_key(ldns_key *k, RSA *r)
 {
-	k->_key.rsa = r;
+	EVP_PKEY *key = EVP_PKEY_new();
+	EVP_PKEY_set1_RSA(key, r);
+	k->_key.key = key;
 }
 
 void
 ldns_key_set_dsa_key(ldns_key *k, DSA *d)
 {
-	k->_key.dsa  = d;
+	EVP_PKEY *key = EVP_PKEY_new();
+	EVP_PKEY_set1_DSA(key, d);
+	k->_key.key  = key;
 }
 
 void
@@ -592,16 +705,39 @@ ldns_key_algorithm(const ldns_key *k)
 	return k->_alg;
 }
 
+void
+ldns_key_set_use(ldns_key *k, bool v)
+{
+	if (k) {
+		k->_use = v;
+	}
+}
+
+bool
+ldns_key_use(const ldns_key *k)
+{
+	if (k) {
+		return k->_use;
+	}
+	return false;
+}
+
+EVP_PKEY *
+ldns_key_evp_key(const ldns_key *k)
+{
+	return k->_key.key;
+}
+
 RSA *
 ldns_key_rsa_key(const ldns_key *k)
 {
-	return k->_key.rsa;
+	return EVP_PKEY_get1_RSA(k->_key.key);
 }
 
 DSA *
 ldns_key_dsa_key(const ldns_key *k)
 {
-	return k->_key.dsa;
+	return EVP_PKEY_get1_DSA(k->_key.key);
 }
 
 unsigned char *
@@ -653,6 +789,16 @@ ldns_key_pubkey_owner(const ldns_key *k)
 }
 
 /* write */
+void
+ldns_key_list_set_use(ldns_key_list *keys, bool v)
+{
+	size_t i;
+
+	for (i = 0; i < ldns_key_list_key_count(keys); i++) {
+		ldns_key_set_use(ldns_key_list_key(keys, i), v);
+	}
+}
+
 void            
 ldns_key_list_set_key_count(ldns_key_list *key, size_t count)
 {
@@ -755,6 +901,8 @@ ldns_key_dsa2bin(unsigned char *data, DSA *k, uint16_t *size)
 	memcpy(data, &T, 1);
 
 	if (T > 8) {
+		fprintf(stderr, "DSA key with T > 8 (ie. > 1024 bits)");
+		fprintf(stderr, " not implemented\n");
 		return false;
 	}
 
@@ -764,7 +912,7 @@ ldns_key_dsa2bin(unsigned char *data, DSA *k, uint16_t *size)
 	BN_bn2bin(k->p, data + 21 ); 		/* offset octects */
 	BN_bn2bin(k->g, data + 21 + *size); 	/* offset octets */
 	BN_bn2bin(k->pub_key, data + 21 + *size + *size); /* offset octets */
-	*size = 20 + (*size * 3);
+	*size = 21 + (*size * 3);
 	return true;
 }
 
@@ -779,7 +927,9 @@ ldns_key2rr(const ldns_key *k)
 	ldns_rr *pubkey;
 	ldns_rdf *keybin;
 	unsigned char *bin;
-	uint16_t size;
+	uint16_t size = 0;
+	RSA *rsa = NULL;
+	DSA *dsa = NULL;
 
 	pubkey = ldns_rr_new();
 	if (!k) {
@@ -793,6 +943,8 @@ ldns_key2rr(const ldns_key *k)
 
 	switch (ldns_key_algorithm(k)) {
 	case LDNS_SIGN_HMACMD5:
+	case LDNS_SIGN_HMACSHA1:
+	case LDNS_SIGN_HMACSHA256:
 		ldns_rr_set_type(pubkey, LDNS_RR_TYPE_KEY);
         	break;
 	default:
@@ -801,51 +953,71 @@ ldns_key2rr(const ldns_key *k)
         }
 	/* zero-th rdf - flags */
 	ldns_rr_push_rdf(pubkey,
-			ldns_native2rdf_int16(LDNS_RDF_TYPE_INT16, 
+			ldns_native2rdf_int16(LDNS_RDF_TYPE_INT16,
 				ldns_key_flags(k)));
 	/* first - proto */
-	ldns_rr_push_rdf(pubkey, 
+	ldns_rr_push_rdf(pubkey,
 			ldns_native2rdf_int8(LDNS_RDF_TYPE_INT8, LDNS_DNSSEC_KEYPROTO));
-	
+
 	if (ldns_key_pubkey_owner(k)) {
 		ldns_rr_set_owner(pubkey, ldns_rdf_clone(ldns_key_pubkey_owner(k)));
 	}
-	
+
 	/* third - da algorithm */
 	switch(ldns_key_algorithm(k)) {
-		case LDNS_SIGN_RSAMD5:
+		case LDNS_RSAMD5:
+		case LDNS_RSASHA1:
+		case LDNS_RSASHA1_NSEC3:
+		case LDNS_RSASHA256:
+		case LDNS_RSASHA256_NSEC3:
+		case LDNS_RSASHA512:
+		case LDNS_RSASHA512_NSEC3:
 			ldns_rr_push_rdf(pubkey,
-					ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, LDNS_RSAMD5));
-			if (!ldns_key_rsa2bin(bin, ldns_key_rsa_key(k), &size)) {
-				return NULL;
-			}
-			break;
-		case LDNS_SIGN_RSASHA1:
-			ldns_rr_push_rdf(pubkey,
-					ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, LDNS_RSASHA1));
-			if (!ldns_key_rsa2bin(bin, ldns_key_rsa_key(k), &size)) {
+						  ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, ldns_key_algorithm(k)));
+			rsa =  ldns_key_rsa_key(k);
+			if (!rsa) {
 				return NULL;
+			} else {
+				if (!ldns_key_rsa2bin(bin, rsa, &size)) {
+					return NULL;
+				}
+				RSA_free(rsa);
 			}
+			size++;
 			break;
 		case LDNS_SIGN_DSA:
 			ldns_rr_push_rdf(pubkey,
 					ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, LDNS_DSA));
+			dsa = ldns_key_dsa_key(k);
+			if (!dsa) {
+				return NULL;
+			} else {
+				if (!ldns_key_dsa2bin(bin, dsa, &size)) {
+					return NULL;
+				}
+				DSA_free(dsa);
+			}
+			break;
+		case LDNS_DSA_NSEC3:
+			ldns_rr_push_rdf(pubkey,
+					ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, LDNS_DSA_NSEC3));
 			if (!ldns_key_dsa2bin(bin, ldns_key_dsa_key(k), &size)) {
 				return NULL;
 			}
 			break;
 		case LDNS_SIGN_HMACMD5:
+		case LDNS_SIGN_HMACSHA1:
+		case LDNS_SIGN_HMACSHA256:
 			/* tja */
 			ldns_rr_push_rdf(pubkey,
-                                         ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG, LDNS_SIGN_HMACMD5));
+                                        ldns_native2rdf_int8(LDNS_RDF_TYPE_ALG,
+						ldns_key_algorithm(k)));
                         size = ldns_key_hmac_size(k);
-                        bin = LDNS_XREALLOC(bin, unsigned char, size);
-                        memcpy(bin, ldns_key_hmac_key(k), size);
+			bin = LDNS_XREALLOC(bin, unsigned char, size);
+			memcpy(bin, ldns_key_hmac_key(k), size);
 			break;
 	}
 	/* fourth the key bin material */
-	/* MIEK, not sure about this +1. I've re-added it--needs checking */
-	/* TODO: and i've removed it again, it's certainly wrong for HMAC */
 	keybin = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, size, bin);
 	LDNS_FREE(bin);
 	ldns_rr_push_rdf(pubkey, keybin);
@@ -864,24 +1036,12 @@ ldns_key_deep_free(ldns_key *key)
 	if (ldns_key_pubkey_owner(key)) {
 		ldns_rdf_deep_free(ldns_key_pubkey_owner(key));
 	}
-	switch(ldns_key_algorithm(key)) {
-	case LDNS_SIGN_RSASHA1:
-	case LDNS_SIGN_RSAMD5:
-		if (ldns_key_rsa_key(key)) {
-			RSA_free(ldns_key_rsa_key(key));
-		}
-		break;
-	case LDNS_SIGN_DSA:
-		if (ldns_key_dsa_key(key)) {
-			DSA_free(ldns_key_dsa_key(key));
-		}
-		break;
-	case LDNS_SIGN_HMACMD5:
-		break;
+	if (ldns_key_evp_key(key)) {
+		EVP_PKEY_free(ldns_key_evp_key(key));
 	}
 	if (ldns_key_hmac_key(key)) {
 		free(ldns_key_hmac_key(key));
-	}	
+	}
 	LDNS_FREE(key);
 }
 
@@ -895,4 +1055,48 @@ ldns_key_list_free(ldns_key_list *key_list)
 	LDNS_FREE(key_list->_keys);
 	LDNS_FREE(key_list);
 }
+
+ldns_rr *
+ldns_read_anchor_file(const char *filename)
+{
+	FILE *fp;
+	/*char line[LDNS_MAX_PACKETLEN];*/
+	char *line = LDNS_XMALLOC(char, LDNS_MAX_PACKETLEN);
+	int c;
+	size_t i = 0;
+	ldns_rr *r;
+	ldns_status status;
+
+	fp = fopen(filename, "r");
+	if (!fp) {
+		fprintf(stderr, "Unable to open %s: %s\n", filename, strerror(errno));
+		LDNS_FREE(line);
+		return NULL;
+	}
+	
+	while ((c = fgetc(fp)) && i < LDNS_MAX_PACKETLEN && c != EOF) {
+		line[i] = c;
+		i++;
+	}
+	line[i] = '\0';
+	
+	fclose(fp);
+	
+	if (i <= 0) {
+		fprintf(stderr, "nothing read from %s", filename);
+		LDNS_FREE(line);
+		return NULL;
+	} else {
+		status = ldns_rr_new_frm_str(&r, line, 0, NULL, NULL);
+		if (status == LDNS_STATUS_OK && (ldns_rr_get_type(r) == LDNS_RR_TYPE_DNSKEY || ldns_rr_get_type(r) == LDNS_RR_TYPE_DS)) {
+			LDNS_FREE(line);
+			return r;
+		} else {
+			fprintf(stderr, "Error creating DNSKEY or DS rr from %s: %s\n", filename, ldns_get_errorstr_by_id(status));
+			LDNS_FREE(line);
+			return NULL;
+		}
+	}
+}
+
 #endif /* HAVE_SSL */
