@@ -44,6 +44,12 @@ ldns_resolver_retrans(const ldns_resolver *r)
 	return r->_retrans;
 }
 
+bool
+ldns_resolver_fallback(const ldns_resolver *r)
+{
+	return r->_fallback;
+}
+
 uint8_t
 ldns_resolver_ip6(const ldns_resolver *r)
 {
@@ -116,6 +122,39 @@ ldns_resolver_dnssec_cd(const ldns_resolver *r)
 	return r->_dnssec_cd;
 }
 
+ldns_rr_list *
+ldns_resolver_dnssec_anchors(const ldns_resolver *r)
+{
+    return r->_dnssec_anchors;
+}
+
+bool
+ldns_resolver_trusted_key(const ldns_resolver *r, ldns_rr_list * keys, ldns_rr_list * trusted_keys)
+{
+  size_t i;
+  bool result = false;
+  
+  ldns_rr_list * trust_anchors;
+  ldns_rr * cur_rr;
+
+  if (!r || !keys) { return false; }
+
+  trust_anchors = ldns_resolver_dnssec_anchors(r);
+
+  if (!trust_anchors) { return false; }
+
+  for (i = 0; i < ldns_rr_list_rr_count(keys); i++) {
+
+    cur_rr = ldns_rr_list_rr(keys, i);
+    if (ldns_rr_list_contains_rr(trust_anchors, cur_rr)) {      
+      if (trusted_keys) { ldns_rr_list_push_rr(trusted_keys, cur_rr); }
+      result = true;
+    }
+  }
+
+  return result;
+}
+
 bool
 ldns_resolver_igntc(const ldns_resolver *r)
 {
@@ -319,6 +358,29 @@ ldns_resolver_set_dnssec_cd(ldns_resolver *r, bool d)
 	r->_dnssec_cd = d;
 }
 
+void
+ldns_resolver_set_dnssec_anchors(ldns_resolver *r, ldns_rr_list * l)
+{
+  r->_dnssec_anchors = l;
+}
+
+ldns_status
+ldns_resolver_push_dnssec_anchor(ldns_resolver *r, ldns_rr *rr)
+{
+  ldns_rr_list * trust_anchors;
+
+  if ((!rr) || (ldns_rr_get_type(rr) != LDNS_RR_TYPE_DNSKEY)) {
+    return LDNS_STATUS_ERR;
+  }
+
+  if (!(trust_anchors = ldns_resolver_dnssec_anchors(r))) { /* Initialize */
+    trust_anchors = ldns_rr_list_new();
+    ldns_resolver_set_dnssec_anchors(r, trust_anchors);
+  }
+
+  return (ldns_rr_list_push_rr(trust_anchors, ldns_rr_clone(rr))) ? LDNS_STATUS_OK : LDNS_STATUS_ERR;
+}
+
 void
 ldns_resolver_set_igntc(ldns_resolver *r, bool i)
 {
@@ -379,6 +441,12 @@ ldns_resolver_set_retrans(ldns_resolver *r, uint8_t retrans)
 	r->_retrans = retrans;
 }
 
+void
+ldns_resolver_set_fallback(ldns_resolver *r, bool fallback)
+{
+	r->_fallback = fallback;
+}
+
 void
 ldns_resolver_set_nameservers(ldns_resolver *r, ldns_rdf **n)
 {
@@ -519,10 +587,12 @@ ldns_resolver_new(void)
 	ldns_resolver_set_defnames(r, false);
 	ldns_resolver_set_retry(r, 3);
 	ldns_resolver_set_retrans(r, 2);
+	ldns_resolver_set_fallback(r, true);
 	ldns_resolver_set_fail(r, false);
 	ldns_resolver_set_edns_udp_size(r, 0);
 	ldns_resolver_set_dnssec(r, false);
 	ldns_resolver_set_dnssec_cd(r, false);
+	ldns_resolver_set_dnssec_anchors(r, NULL);
 	ldns_resolver_set_ip6(r, LDNS_RESOLV_INETANY);
 
 	/* randomize the nameserver to be queried
@@ -561,6 +631,9 @@ ldns_resolver_new_frm_fp_l(ldns_resolver **res, FILE *fp, int *line_nr)
 	int8_t expect;
 	uint8_t i;
 	ldns_rdf *tmp;
+#ifdef HAVE_SSL
+	ldns_rr *tmp_rr;
+#endif
 	ssize_t gtr;
 	ldns_buffer *b;
 
@@ -578,6 +651,7 @@ ldns_resolver_new_frm_fp_l(ldns_resolver **res, FILE *fp, int *line_nr)
 	/* these two are read but not used atm TODO */
 	keyword[LDNS_RESOLV_SORTLIST] = "sortlist";
 	keyword[LDNS_RESOLV_OPTIONS] = "options";
+	keyword[LDNS_RESOLV_ANCHOR] = "anchor";
 	expect = LDNS_RESOLV_KEYWORD;
 
 	r = ldns_resolver_new();
@@ -610,15 +684,13 @@ ldns_resolver_new_frm_fp_l(ldns_resolver **res, FILE *fp, int *line_nr)
 							break;
 						}
 					}
-					if (word[0] == '#') {
-						/* skip the rest of the line*/
-						gtr = ldns_fget_token_l(fp, word, LDNS_PARSE_SKIP_SPACE, 0, line_nr);
-						continue;
-					}
 					/* no keyword recognized */
 					if (expect == LDNS_RESOLV_KEYWORD) {
+						/* skip line */
+						/*
 						ldns_resolver_deep_free(r);
 						return LDNS_STATUS_SYNTAX_KEYWORD_ERR;
+						*/
 					}
 				}
 				break;
@@ -691,6 +763,20 @@ ldns_resolver_new_frm_fp_l(ldns_resolver **res, FILE *fp, int *line_nr)
 				/* options not implemented atm */
 				expect = LDNS_RESOLV_KEYWORD;
 				break;
+			case LDNS_RESOLV_ANCHOR:
+				/* a file containing a DNSSEC trust anchor */
+				gtr = ldns_fget_token_l(fp, word, LDNS_PARSE_NORMAL, 0, line_nr);
+				if (gtr == 0) {
+					return LDNS_STATUS_SYNTAX_MISSING_VALUE_ERR;
+				}
+
+#ifdef HAVE_SSL
+				tmp_rr = ldns_read_anchor_file(word);
+				(void) ldns_resolver_push_dnssec_anchor(r, tmp_rr);
+				ldns_rr_free(tmp_rr);
+#endif
+				expect = LDNS_RESOLV_KEYWORD;
+				break;
 		}
 	}
 	
@@ -770,6 +856,9 @@ ldns_resolver_deep_free(ldns_resolver *res)
 		if (res->_rtt) {
 			LDNS_FREE(res->_rtt);
 		}
+		if (res->_dnssec_anchors) {
+			ldns_rr_list_deep_free(res->_dnssec_anchors);
+		}
 		LDNS_FREE(res);
 	}
 }
@@ -852,13 +941,15 @@ ldns_resolver_query(const ldns_resolver *r, const ldns_rdf *name, ldns_rr_type t
 	}
 	status = ldns_resolver_send(&pkt, (ldns_resolver *)r, newname, type, class, 
 			flags);
+
 	ldns_rdf_free(newname);
+
 	return pkt;
 }
 
 ldns_status
-ldns_resolver_send_pkt(ldns_pkt **answer,const ldns_resolver *r, 
-		const ldns_pkt *query_pkt)
+ldns_resolver_send_pkt(ldns_pkt **answer, ldns_resolver *r, 
+				   ldns_pkt *query_pkt)
 {
 	ldns_pkt *answer_pkt = NULL;
 	ldns_status stat = LDNS_STATUS_OK;
@@ -869,8 +960,31 @@ ldns_resolver_send_pkt(ldns_pkt **answer,const ldns_resolver *r,
 			ldns_pkt_free(answer_pkt);
 			answer_pkt = NULL;
 		}
-	}
+	} else {
 	
+		/* if tc=1 fall back to EDNS and/or TCP */
+		/* check for tcp first (otherwise we don't care about tc=1) */
+		if (!ldns_resolver_usevc(r) && ldns_resolver_fallback(r)) {
+			if (ldns_pkt_tc(answer_pkt)) {
+				/* was EDNS0 set? */
+				if (ldns_pkt_edns_udp_size(query_pkt) == 0) {
+					ldns_pkt_set_edns_udp_size(query_pkt, 4096);
+					ldns_pkt_free(answer_pkt);
+					stat = ldns_send(&answer_pkt, r, query_pkt);
+				}
+				/* either way, if it is still truncated, use TCP */
+				if (stat != LDNS_STATUS_OK ||
+				    ldns_pkt_tc(answer_pkt)) {
+					ldns_resolver_set_usevc(r, true);
+					ldns_pkt_free(answer_pkt);
+					stat = ldns_send(&answer_pkt, r, query_pkt);
+					ldns_resolver_set_usevc(r, false);
+				}
+			}
+		}
+	}
+
+
 	if (answer) {
 		*answer = answer_pkt;
 	}
@@ -896,7 +1010,9 @@ ldns_resolver_prepare_query_pkt(ldns_pkt **query_pkt, ldns_resolver *r,
 			ldns_resolver_set_edns_udp_size(r, 4096);
 		}
 		ldns_pkt_set_edns_do(*query_pkt, true);
-		ldns_pkt_set_cd(*query_pkt, ldns_resolver_dnssec_cd(r));
+		if (ldns_resolver_dnssec_cd(r) || (flags & LDNS_CD)) {
+			ldns_pkt_set_cd(*query_pkt, true);
+		}
 	}
 
 	/* transfer the udp_edns_size from the resolver to the packet */
@@ -1076,3 +1192,4 @@ ldns_resolver_nameservers_randomize(ldns_resolver *r)
 	}
 	ldns_resolver_set_nameservers(r, ns);
 }
+
