@@ -81,7 +81,8 @@ ldns_rr_free(ldns_rr *rr)
 }
 
 /* 
- * extra spaces are allowed
+ * trailing spaces are allowed
+ * leading spaces are not allowed
  * allow ttl to be optional
  * class is optional too
  * if ttl is missing, and default_ttl is 0, use DEF_TTL
@@ -93,9 +94,10 @@ ldns_rr_free(ldns_rr *rr)
  * or
  * miek.nl. IN MX 10 elektron.atoom.net
  */
-ldns_status
-ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns_rdf *origin, 
-		ldns_rdf **prev)
+static ldns_status
+ldns_rr_new_frm_str_internal(ldns_rr **newrr, const char *str,
+                             uint32_t default_ttl, ldns_rdf *origin, 
+		             ldns_rdf **prev, bool question)
 {
 	ldns_rr *new;
 	const ldns_rr_descriptor *desc;
@@ -141,7 +143,7 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 	rd_buf = LDNS_MALLOC(ldns_buffer);
 	rd = LDNS_XMALLOC(char, LDNS_MAX_RDFLEN);
 	b64 = LDNS_XMALLOC(char, LDNS_MAX_RDFLEN);
-	if (!new || !owner || !ttl || !clas || !rdata || !rr_buf || !rd_buf || !rd | !b64) {
+	if (!new || !owner || !ttl || !clas || !rdata || !rr_buf || !rd_buf || !rd || !b64 ) {
 		return LDNS_STATUS_MEM_ERR;
 	}
 	r_cnt = 0;
@@ -178,7 +180,7 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 	}
 	ttl_val = (uint32_t) strtol(ttl, NULL, 10);
 
-	if (strlen(ttl) > 0 && !isdigit(ttl[0])) {
+	if (strlen(ttl) > 0 && !isdigit((int) ttl[0])) {
 		/* ah, it's not there or something */
 		if (default_ttl == 0) {
 			ttl_val = LDNS_DEFAULT_TTL;
@@ -346,7 +348,9 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 				quoted = false;
 				/* if type = B64, the field may contain spaces */
 				if (ldns_rr_descriptor_field_type(desc, 
-							r_cnt) == LDNS_RDF_TYPE_B64 ||
+					    r_cnt) == LDNS_RDF_TYPE_B64 ||
+				    ldns_rr_descriptor_field_type(desc, 
+					    r_cnt) == LDNS_RDF_TYPE_HEX ||
 				    ldns_rr_descriptor_field_type(desc, 
 					    r_cnt) == LDNS_RDF_TYPE_LOC ||
 				    ldns_rr_descriptor_field_type(desc, 
@@ -379,8 +383,15 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 
 				/* because number of fields can be variable, we can't
 				   rely on _maximum() only */
+				/* skip spaces */
+				while (ldns_buffer_position(rd_buf) < ldns_buffer_limit(rd_buf) &&
+					*(ldns_buffer_current(rd_buf)) == ' ' 
+				      ) {
+					ldns_buffer_skip(rd_buf, 1);
+				}
+
 				if ((c = ldns_bget_token(rd_buf, rd, delimiters, 
-									LDNS_MAX_RDFLEN)) != -1) {
+							LDNS_MAX_RDFLEN)) != -1) {
 					/* hmmz, rfc3597 specifies that any type can be represented with
 					 * \# method, which can contain spaces...
 					 * it does specify size though... 
@@ -432,6 +443,7 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 					} else {
 						/* Normal RR */
 						switch(ldns_rr_descriptor_field_type(desc, r_cnt)) {
+						case LDNS_RDF_TYPE_HEX:
 						case LDNS_RDF_TYPE_B64:
 							/* can have spaces, and will always be the last 
 							 * record of the rrdata. Read in the rest */
@@ -490,7 +502,6 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 							ldns_rr_free(new);
 							return LDNS_STATUS_SYNTAX_RDATA_ERR;
 						}
-						
 					}
 					if (quoted) {
 						if (ldns_buffer_available(rd_buf, 1)) {
@@ -510,12 +521,42 @@ ldns_rr_new_frm_str(ldns_rr **newrr, const char *str, uint32_t default_ttl, ldns
 	ldns_buffer_free(rr_buf);
 	LDNS_FREE(rdata);
 
+	if (!question && desc && ldns_rr_rd_count(new) < r_min) {
+		ldns_rr_free(new);
+		return LDNS_STATUS_SYNTAX_MISSING_VALUE_ERR;
+	}
+
 	if (newrr) {
 		*newrr = new;
 	}
 	return LDNS_STATUS_OK;
 }
 
+ldns_status
+ldns_rr_new_frm_str(ldns_rr **newrr, const char *str,
+                    uint32_t default_ttl, ldns_rdf *origin, 
+		    ldns_rdf **prev)
+{
+	return ldns_rr_new_frm_str_internal(newrr,
+	                                    str,
+	                                    default_ttl,
+	                                    origin,
+	                                    prev,
+	                                    false);
+}
+
+ldns_status
+ldns_rr_new_question_frm_str(ldns_rr **newrr, const char *str,
+                             ldns_rdf *origin, ldns_rdf **prev)
+{
+	return ldns_rr_new_frm_str_internal(newrr,
+	                                    str,
+	                                    0,
+	                                    origin,
+	                                    prev,
+	                                    true);
+}
+
 ldns_status
 ldns_rr_new_frm_fp(ldns_rr **newrr, FILE *fp, uint32_t *ttl, ldns_rdf **origin, ldns_rdf **prev)
 {
@@ -525,7 +566,7 @@ ldns_rr_new_frm_fp(ldns_rr **newrr, FILE *fp, uint32_t *ttl, ldns_rdf **origin,
 ldns_status
 ldns_rr_new_frm_fp_l(ldns_rr **newrr, FILE *fp, uint32_t *default_ttl, ldns_rdf **origin, ldns_rdf **prev, int *line_nr)
 {
-        char *line;
+	char *line;
 	const char *endptr;  /* unused */
 	ldns_rr *rr;
 	char *keyword;
@@ -541,20 +582,20 @@ ldns_rr_new_frm_fp_l(ldns_rr **newrr, FILE *fp, uint32_t *default_ttl, ldns_rdf
 		ttl = 0;
 	}
 
-        line = LDNS_XMALLOC(char, LDNS_MAX_LINELEN + 1);
-        if (!line) {
-                return LDNS_STATUS_MEM_ERR;
-        }
+	line = LDNS_XMALLOC(char, LDNS_MAX_LINELEN + 1);
+	if (!line) {
+		return LDNS_STATUS_MEM_ERR;
+	}
 
-        /* read an entire line in from the file */
-        if ((size = ldns_fget_token_l(fp, line, LDNS_PARSE_SKIP_SPACE, LDNS_MAX_LINELEN, line_nr)) == -1) {
+	/* read an entire line in from the file */
+	if ((size = ldns_fget_token_l(fp, line, LDNS_PARSE_SKIP_SPACE, LDNS_MAX_LINELEN, line_nr)) == -1) {
 		LDNS_FREE(line);
 		/* if last line was empty, we are now at feof, which is not
 		 * always a parse error (happens when for instance last line
 		 * was a comment)
 		 */
-                return LDNS_STATUS_SYNTAX_ERR;
-        }
+		return LDNS_STATUS_SYNTAX_ERR;
+	}
 
 	/* we can have the situation, where we've read ok, but still got
 	 * no bytes to play with, in this case size is 0 
@@ -576,7 +617,7 @@ ldns_rr_new_frm_fp_l(ldns_rr **newrr, FILE *fp, uint32_t *default_ttl, ldns_rdf
 			return LDNS_STATUS_SYNTAX_DNAME_ERR;
 		}
 		*origin = tmp;
-		s = LDNS_STATUS_SYNTAX_ORIGIN;		
+		s = LDNS_STATUS_SYNTAX_ORIGIN;
 	} else if ((keyword = strstr(line, "$TTL "))) {
 		if (default_ttl) {
 			*default_ttl = ldns_str2period(keyword + 5, &endptr);
@@ -681,7 +722,7 @@ ldns_rr_pop_rdf(ldns_rr *rr)
 		return NULL;
 	}
 
-	pop = rr->_rdata_fields[rd_count];
+	pop = rr->_rdata_fields[rd_count - 1];
 	
 	/* shrink the array */
 	rr->_rdata_fields = LDNS_XREALLOC(
@@ -1852,13 +1893,13 @@ static ldns_rr_descriptor rdata_field_descriptors[] = {
 	/* 46 */
 	{LDNS_RR_TYPE_RRSIG, "RRSIG", 9, 9, type_rrsig_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 1 },
 	/* 47 */
-	{LDNS_RR_TYPE_NSEC, "NSEC", 2, 2, type_nsec_wireformat, LDNS_RDF_TYPE_NSEC, LDNS_RR_NO_COMPRESS, 1 },
+	{LDNS_RR_TYPE_NSEC, "NSEC", 1, 2, type_nsec_wireformat, LDNS_RDF_TYPE_NSEC, LDNS_RR_NO_COMPRESS, 1 },
 	/* 48 */
 	{LDNS_RR_TYPE_DNSKEY, "DNSKEY", 4, 4, type_dnskey_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 	/* 49 */
 {LDNS_RR_TYPE_DHCID, "DHCID", 1, 1, type_dhcid_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 	/* 50 */
-{LDNS_RR_TYPE_NSEC3, "NSEC3", 6, 6, type_nsec3_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
+	{LDNS_RR_TYPE_NSEC3, "NSEC3", 5, 6, type_nsec3_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 	/* 51 */
 {LDNS_RR_TYPE_NSEC3PARAMS, "NSEC3PARAM", 4, 4, type_nsec3params_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 	/* 52 */
@@ -2060,7 +2101,9 @@ static ldns_rr_descriptor rdata_field_descriptors[] = {
 {LDNS_RR_TYPE_NULL, "TYPE247", 1, 1, type_0_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 {LDNS_RR_TYPE_NULL, "TYPE248", 1, 1, type_0_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
 {LDNS_RR_TYPE_NULL, "TYPE249", 1, 1, type_0_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
-{LDNS_RR_TYPE_TSIG, "TSIG", 8, 9, type_tsig_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 }
+{LDNS_RR_TYPE_TSIG, "TSIG", 8, 9, type_tsig_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 },
+/* split in array, no longer contiguous */
+{LDNS_RR_TYPE_DLV, "DLV", 4, 4, type_ds_wireformat, LDNS_RDF_TYPE_NONE, LDNS_RR_NO_COMPRESS, 0 }
 };
 /** \endcond */
 
@@ -2078,7 +2121,7 @@ ldns_rr_descript(uint16_t type)
 	if (type <= LDNS_RDATA_FIELD_DESCRIPTORS_COMMON) {
 		return &rdata_field_descriptors[type];
 	} else {
-	        /* TODO: tmp until type code */
+		/* because not all array index equals type code */
 		for (i = LDNS_RDATA_FIELD_DESCRIPTORS_COMMON;
 		     i < LDNS_RDATA_FIELD_DESCRIPTORS_COUNT;
 		     i++) {
@@ -2148,7 +2191,8 @@ ldns_get_rr_type_by_name(const char *name)
 		if(desc_name &&
 		   strlen(name) == strlen(desc_name) &&
 		   strncasecmp(name, desc_name, strlen(desc_name)) == 0) {
-			return i;
+			/* because not all array index equals type code */
+			return desc->_type;
 		}
 	}
 	
@@ -2204,3 +2248,24 @@ ldns_rdf2rr_type(const ldns_rdf *rd)
         r = (ldns_rr_type) ldns_rdf2native_int16(rd);
         return r;
 }
+
+ldns_rr_type
+ldns_rr_list_type(const ldns_rr_list *rr_list)
+{
+	if (rr_list && ldns_rr_list_rr_count(rr_list) > 0) {
+		return ldns_rr_get_type(ldns_rr_list_rr(rr_list, 0));
+	} else {
+		return 0;
+	}
+}
+
+ldns_rdf *
+ldns_rr_list_owner(const ldns_rr_list *rr_list)
+{
+	if (rr_list && ldns_rr_list_rr_count(rr_list) > 0) {
+		return ldns_rr_owner(ldns_rr_list_rr(rr_list, 0));
+	} else {
+		return NULL;
+	}
+}
+
