@@ -138,7 +138,7 @@ ldns_str2rdf_nsec3_salt(ldns_rdf **rd, const char *salt_str)
 	
 	salt = LDNS_XMALLOC(uint8_t, salt_length / 2);
 	for (c = 0; c < salt_length; c += 2) {
-		if (isxdigit(salt_str[c]) && isxdigit(salt_str[c+1])) {
+		if (isxdigit((int) salt_str[c]) && isxdigit((int) salt_str[c+1])) {
 			salt[c/2] = (uint8_t) ldns_hexdigit_to_int(salt_str[c]) * 16 +
 					  ldns_hexdigit_to_int(salt_str[c+1]);
 		} else {
@@ -190,7 +190,11 @@ ldns_str2rdf_int32(ldns_rdf **rd, const char *longstr)
 	if(*end != 0) {
 		LDNS_FREE(r);
 		return LDNS_STATUS_ERR;
-        } else {
+     } else {
+		if (errno == ERANGE) {
+			LDNS_FREE(r);
+			return LDNS_STATUS_SYNTAX_INTEGER_OVERFLOW;
+		}
 		memcpy(r, &l, sizeof(uint32_t));
 		*rd = ldns_rdf_new_frm_data(
 			LDNS_RDF_TYPE_INT32, sizeof(uint32_t), r);
@@ -236,7 +240,7 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 	
 	len = strlen((char*)str);
 	/* octet representation can make strings a lot longer than actual length */
-	if (len > LDNS_MAX_DOMAINLEN * 3) {
+	if (len > LDNS_MAX_DOMAINLEN * 4) {
 		return LDNS_STATUS_DOMAINNAME_OVERFLOW;
 	}
 	if (0 == len) {
@@ -260,6 +264,9 @@ ldns_str2rdf_dname(ldns_rdf **d, const char *str)
 	pq = buf;
 	label_len = 0;
 	for (s = p = (uint8_t *) str; *s; s++, q++) {
+		if (q > buf + LDNS_MAX_DOMAINLEN) {
+			return LDNS_STATUS_DOMAINNAME_OVERFLOW;
+		}	
 		*q = 0;
 		switch (*s) {
 		case '.':
@@ -354,9 +361,9 @@ ldns_str2rdf_str(ldns_rdf **rd, const char *str)
 	for (str_i = 0; str_i < strlen(str); str_i++) {
 		if (str[str_i] == '\\') {
 			if(str_i + 3 < strlen(str) && 
-			   isdigit(str[str_i + 1]) &&
-			   isdigit(str[str_i + 2]) &&
-			   isdigit(str[str_i + 3])) {
+			   isdigit((int) str[str_i + 1]) &&
+			   isdigit((int) str[str_i + 2]) &&
+			   isdigit((int) str[str_i + 3])) {
 				val = (uint8_t) ldns_hexdigit_to_int((char) str[str_i + 1]) * 100 +
 				                ldns_hexdigit_to_int((char) str[str_i + 2]) * 10 +
 				                ldns_hexdigit_to_int((char) str[str_i + 3]);
@@ -475,10 +482,10 @@ ldns_str2rdf_b64(ldns_rdf **rd, const char *str)
 	uint8_t *buffer;
 	int16_t i;
 	
-	buffer = LDNS_XMALLOC(uint8_t, b64_ntop_calculate_size(strlen(str)));
+	buffer = LDNS_XMALLOC(uint8_t, ldns_b64_ntop_calculate_size(strlen(str)));
 	
-	i = (uint16_t)b64_pton((const char*)str, buffer, 
-	                        b64_ntop_calculate_size(strlen(str)));
+	i = (uint16_t)ldns_b64_pton((const char*)str, buffer, 
+						   ldns_b64_ntop_calculate_size(strlen(str)));
 	if (-1 == i) {
 		LDNS_FREE(buffer);
 		return LDNS_STATUS_INVALID_B64;
@@ -497,12 +504,12 @@ ldns_str2rdf_b32_ext(ldns_rdf **rd, const char *str)
 	uint8_t *buffer;
 	int i;
 	/* first byte contains length of actual b32 data */
-	uint8_t len = b32_pton_calculate_size(strlen(str));
+	uint8_t len = ldns_b32_pton_calculate_size(strlen(str));
 	buffer = LDNS_XMALLOC(uint8_t, len + 1);
 	buffer[0] = len;
 	
-	i = b32_pton_extended_hex((const char*)str, strlen(str), buffer + 1, 
-                                  b32_ntop_calculate_size(strlen(str)));
+	i = ldns_b32_pton_extended_hex((const char*)str, strlen(str), buffer + 1, 
+							 ldns_b32_ntop_calculate_size(strlen(str)));
 	if (i < 0) {
 		return LDNS_STATUS_INVALID_B32_EXT;
 	} else {
@@ -523,30 +530,33 @@ ldns_str2rdf_hex(ldns_rdf **rd, const char *str)
 
 	len = strlen(str);
 
-        if (len % 2 != 0) {
-                return LDNS_STATUS_INVALID_HEX;
-        } else if (len > LDNS_MAX_RDFLEN * 2) {
+	if (len > LDNS_MAX_RDFLEN * 2) {
 		return LDNS_STATUS_LABEL_OVERFLOW;
         } else {
-		t = LDNS_XMALLOC(uint8_t, (len / 2));
+		t = LDNS_XMALLOC(uint8_t, (len / 2) + 1);
 		t_orig = t;
                 /* Now process octet by octet... */
                 while (*str) {
-                        *t = 0;
-                        for (i = 16; i >= 1; i -= 15) {
-                                if (isxdigit(*str)) {
-                                        *t += ldns_hexdigit_to_int(*str) * i;
-                                } else {
-					/* error or be lenient and skip? */
-                                        /*return LDNS_STATUS_ERR;*/
-                                }
-                                ++str;
-                        }
-                        ++t;
+			*t = 0;
+			if (isspace((int) *str)) {
+				str++;
+			} else {
+				for (i = 16; i >= 1; i -= 15) {
+					while (*str && isspace((int) *str)) { str++; }
+					if (*str) {
+						if (isxdigit((int) *str)) {
+							*t += ldns_hexdigit_to_int(*str) * i;
+						} else {
+							return LDNS_STATUS_ERR;
+						}
+						++str;
+					}
+				}
+				++t;
+			}
                 }
-                t = t_orig;
-		*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_HEX, len / 2, t);
-		LDNS_FREE(t);
+		*rd = ldns_rdf_new_frm_data(LDNS_RDF_TYPE_HEX, t - t_orig, t_orig);
+		LDNS_FREE(t_orig);
         }
         return LDNS_STATUS_OK;
 }
@@ -791,17 +801,17 @@ ldns_str2rdf_loc(ldns_rdf **rd, const char *str)
 	char *my_str = (char *) str;
 
 	/* only support version 0 */
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		h = (uint32_t) strtol(my_str, &my_str, 10);
 	} else {
 		return LDNS_STATUS_INVALID_STR;
 	}
 
-	while (isblank(*my_str)) {
+	while (isblank((int) *my_str)) {
 		my_str++;
 	}
 
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		m = (uint32_t) strtol(my_str, &my_str, 10);
 	} else if (*my_str == 'N' || *my_str == 'S') {
 		goto north;
@@ -809,15 +819,15 @@ ldns_str2rdf_loc(ldns_rdf **rd, const char *str)
 		return LDNS_STATUS_INVALID_STR;
 	}
 
-	while (isblank(*my_str)) {
+	while (isblank((int) *my_str)) {
 		my_str++;
 	}
 
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		s = strtod(my_str, &my_str);
 	}
 north:
-	while (isblank(*my_str)) {
+	while (isblank((int) *my_str)) {
 		my_str++;
 	}
 
@@ -847,17 +857,17 @@ north:
 		my_str++;
 	}
 
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		h = (uint32_t) strtol(my_str, &my_str, 10);
 	} else {
 		return LDNS_STATUS_INVALID_STR;
 	}
 
-	while (isblank(*my_str)) {
+	while (isblank((int) *my_str)) {
 		my_str++;
 	}
 
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		m = (uint32_t) strtol(my_str, &my_str, 10);
 	} else if (*my_str == 'E' || *my_str == 'W') {
 		goto east;
@@ -869,7 +879,7 @@ north:
 		my_str++;
 	}
 
-	if (isdigit(*my_str)) {
+	if (isdigit((int) *my_str)) {
 		s = strtod(my_str, &my_str);
 	}
 
